%if 0

lDebug heatshrink depacker

Copyright (C) 2008-2023 E. C. Masloch

Usage of the works is permitted provided that this
instrument is retained with the works, so that any entity
that uses the works is notified of this instrument.

DISCLAIMER: THE WORKS ARE WITHOUT WARRANTY.

%endif

%include "lmacros3.mac"


	numdef STANDALONE,	1

%ifn _STANDALONE
	overridedef DEBUG0,	0
	overridedef COUNTER,	0
%else
	numdef DEBUG0,		0
	numdef COUNTER,		0, 32
%if (_COUNTER - 1) & _COUNTER
 %error COUNTER must be a power of two
%endif


	cpu 8086
	org 256
	addsection CODE, start=256
testprogram:
	mov si, 81h

	cmp sp, stack.top
	jae @F
error:
	mov dx, msg.error
	mov ah, 09h
	int 21h
	mov ax, 4CFFh
	int 21h
	int 20h

@@:
	lodsb
	cmp al, 9
	je @B
	cmp al, 32
	je @B
	cmp al, 13
	je error
	mov dx, si
	dec dx

@@:
	lodsb
	cmp al, 9
	je @F
	cmp al, 32
	je @F
	cmp al, 13
	jne @B
@@:
	mov byte [si - 1], 0

	mov ax, 3D00h | 0_010_0_000b	; RO, DENY WRITE
	int 21h
	jc error
	mov word [handle], ax
	mov word [filebuffer.next], filebuffer
	mov word [filebuffer.tail], filebuffer

	xchg bx, ax
	mov ax, 4202h
	xor cx, cx
	xor dx, dx
	int 21h
	jc error

	push dx
	push ax
	mov ax, 4200h
	xor cx, cx
	xor dx, dx
	int 21h
	jc error

	mov ax, 3C00h
	xor cx, cx
	mov dx, msg.outname
	int 21h
	jc error
	mov word [handle2], ax

	pop dx
	pop cx				; = file size

	mov di, resultbuffer
	push di
	push cx
	mov ax, 0CCCCh
	mov cx, words(resultbuffer.end - resultbuffer)
	rep stosw
	pop cx
	pop di
	mov si, resultbuffer.end

	call depack
	jc error

	push ss
	pop ds
	push ss
	pop es

	mov ah, 3Eh
	mov bx, [handle]
	int 21h

	mov ah, 3Eh
	mov bx, [handle2]
	int 21h

	mov ax, 4C00h
	int 21h
	int 20h


%if _COUNTER
disp_al:
	push dx
	push ax
	xchg ax, dx
	mov ah, 02h
	int 21h
	pop ax
	pop dx
	retn
%endif

get_file_byte:
	mov si, [filebuffer.next]
	cmp si, [filebuffer.tail]
	jb .buffered

	push ax
	push bx
	push cx
	push dx
	mov dx, filebuffer
	mov [filebuffer.next], dx
	mov [filebuffer.tail], dx
	mov ah, 3Fh
	mov bx, [handle]
        mov cx, filebuffer.end - filebuffer

	int 21h
	jc error
	test ax, ax
	jz error
	add ax, dx
	mov word [filebuffer.tail], ax
        mov si, dx
	pop dx
	pop cx
	pop bx
	pop ax

.buffered:
	lodsb
	mov [filebuffer.next], si
	clc
	retn


put_file_data:
	push ax
	push bx
	push dx
	mov dx, si
	mov ah, 40h
	mov bx, [handle2]
	jcxz @F
	int 21h
	jc error
	cmp ax, cx
	jne error
@@:
	pop dx
	pop bx
	pop ax
	retn


	addsection DATA, align=1 follows=CODE
msg:
.error:		ascic "Error!",13,10
.outname:	asciz "output.bin"


	addsection BSSDATA, align=16 nobits follows=DATA
	alignb 16
resultbuffer:
.size equ 4096
		resb .size
.end:
	alignb 2

filebuffer:	resb 256
.end:
	alignb 2
.next:		resw 1
.tail:		resw 1
handle:		resw 1
handle2:	resw 1

stack:		resb 512
.top:


	usesection CODE
%endif	; _STANDALONE


		; INP:	cx:dx = length of source
		;	es:di -> destination
		;	si -> behind end of destination
		; OUT:	NC if success
		;	CY if error,
		;	 bx = ?errordata (if _DEBUG0)
		; CHG:	ax, (bx), cx, dx, es, ds, si, di
		; STT:	UP
depack:
	lframe near
	lenter
 %if _DEBUG0 || _COUNTER
	xor bx, bx
 %endif
 %if _DEBUG0
	lvar word,	errordata
	 push bx
 %endif
%if _COUNTER
	lvar word,	unused_and_counter
	lequ ?unused_and_counter + 1, counter
	 push bx		; initialise counter (high byte) to zero
%endif

	lvar dword,	original_dst
	 push es
	 push di
	lvar dword,	dst
	 push es
	 push di
	lvar dword,	src_remaining
	 push cx
	 push dx

	mov cx, si
	sub cx, di

	lvar word,	dst_length
	 push cx		; push into [bp + ?dst_length]

	mov al, 0
	rep stosb		; prepare for matches before first data byte

@@:
	call read_byte
d0	mov byte [bp + ?errordata], 70h
	jc .error

	mov ah, 0
	test ax, ax
d0	mov byte [bp + ?errordata], 71h
	jz .error
	cmp al, 15
	ja .error

	lvar word,	window_size_bits
	 push ax

	xchg cx, ax
	mov ax, 1
	shl ax, cl
	cmp ax, resultbuffer.size
	ja .error

	call read_byte
d0	mov byte [bp + ?errordata], 72h
	jc .error

	mov ah, 0
	test ax, ax
d0	mov byte [bp + ?errordata], 73h
	jz .error
	cmp ax, word [bp + ?window_size_bits]
	jae .error

	lvar word,	lookahead_size_bits
	 push ax

	xor cx, cx
	lvar word,	low_bit_index_and_high_current_byte
	 push cx

.loop:
%if _COUNTER
	inc byte [bp + ?counter]
	test byte [bp + ?counter], _COUNTER - 1
	jnz @F
	mov al, '.'
	call disp_al
@@:
%endif
	mov dx, word [bp + ?original_dst]
	add dx, word [bp + ?dst_length]	; dx -> end of buffer
	jc .error

	mov cx, 1
	call get_bits
	jnc .notend			; (cx = 0 if jumping)

d0	mov byte [bp + ?errordata], 7Dh
.end_check:
	xor ax, ax
	cmp word [bp + ?src_remaining], ax
	jne .error
	cmp word [bp + ?src_remaining + 2], ax
	jne .error

	db __TEST_IMM8			; (NC)
.error:
	stc

%if _COUNTER
	lahf
	mov al, 13
	call disp_al
	mov al, 10
	call disp_al
	sahf
%endif
d0	mov bx, word [bp + ?errordata]
	lleave code
	lret

.notend:				; (cx = 0)
	test al, al
	jz .notliteral

	mov cl, 8
	call get_bits			; cx = 0
d0	mov byte [bp + ?errordata], 74h
	jc .end_check
	push ax				; on stack: the literal
	 push ss
	 pop ds
	mov si, sp			; ds:si -> data on stack
	xchg ax, cx			; ax = 0
	inc ax				; length = 1
	call copy_data
	pop ax				; (discard)
d0	mov byte [bp + ?errordata], 75h
	jc .error
	jmp .loop

.notliteral:				; (cx = 0)

	mov cl, byte [bp + ?window_size_bits]
					; cx = -w parameter
	call get_bits
d0	mov byte [bp + ?errordata], 76h
	jc .end_check
		; ax = output index
		; cx = 0
	xchg bx, ax
	mov cl, byte [bp + ?lookahead_size_bits]
					; cx = -l parameter
	call get_bits
d0	mov byte [bp + ?errordata], 77h
	jc .end_check
		; bx = output index less 1
		; ax = output count less 1
	inc bx				; = output index
d0	mov byte [bp + ?errordata], 78h
	jz .error

					; ax = length of match less one
	inc ax				; = length of match
	jz .error

	mov cx, bx			; 1 .. 65535

	lds si, [bp + ?dst]
	mov bx, word [bp + ?original_dst]
	neg bx
	add bx, si			; dst - original = length at start
	sub bx, cx			; length start - match distance
	jae @F				; length start >= match distance -->
		; bx = negative length of matchable at end of buffer

	mov si, dx
	add si, bx			; si -> match data
	mov bx, si
	add bx, ax			; -> behind match data (if fits)
	jc .full			; if doesn't fit, copy full -->
	cmp bx, dx			; -> behind is above-or-equal end ?
	; jae .full			; yes -->
	jb .copy_data			; ax has full length from tail of buffer -->

		; if we branch to here,
		;  si + ax >= dx
		;  therefore  ax >= dx - si
.full:
	push ax				; save original length
		; calculate dx - si
	mov ax, dx
	sub ax, si			; -> end minus -> match data
	push ax				; save length
	call copy_data
d0	mov byte [bp + ?errordata], 1
	jc .error
	pop bx				; = done length
	pop ax				; = entire length
	sub ax, bx			; entire length minus done length
	mov si, word [bp + ?original_dst]	; -> start of buffer
	db __TEST_IMM16			; skip sub
@@:
	sub si, cx			; -> into buffer

.copy_data:
	call copy_data			; give ?dst -> dest, ds:si -> source
		; returns: ?dst incremented, ds:si -> after match source
d0	mov byte [bp + ?errordata], 7Bh
	jc .error

	jmp .loop


		; INP:	?dst -> destination,
		;	 destination may overlap the circular buffer boundary.
		;	ds:si -> source
		;	ax = how long the data is (0 is valid),
		;	 always 1 for literals.
		;	 source must fit in circular buffer, except if
		;	 source and destination overlap. in this case the
		;	 source starts in the buffer but may extend past
		;	 the circular buffer boundary.
		;	dx -> end of circular buffer
		;	word [?original_dst] -> start of circular buffer
		; OUT:	?dst incremented
		;	CY if error (buffer too small)
		;	NC if success
		; CHG:	cx, bx, dx, ax, es, di, ds, si
copy_data:
%if 0
	push ax
	push bx
	push cx
	push dx
	mov ax, 4201h
	mov cx, 0
	mov dx, 0
	mov bx, word [handle2]
	int 21h
	nop
	pop dx
	pop cx
	pop bx
	pop ax
%endif

d0	inc byte [bp + ?errordata + 1]

	les di, [bp + ?dst]

	xchg cx, ax			; cx = remaining length
	jcxz .end

	cmp di, dx
	ja .ret_CY			; error
		; literals may branch to .start
	je .start			; ?dst -> at end (put whole data at start)
	add di, cx			; -> behind data chunk (if fits)
	jc .fill			; doesn't fit -->
	cmp di, dx			; fits ?
		; literals never branch to .fill
	ja .fill			; doesn't fit -->
	mov di, word [bp + ?dst]	; restore -> destination
		; ds:si l cx fits in circular buffer
	jmp .mov			; store full up to below-or-equal -> end

.fill:
	mov di, word [bp + ?dst]	; restore -> destination
	mov ax, dx			; -> behind buffer
	sub ax, di			; -> behind buffer minus -> destination
					; = length fitting at end
	sub cx, ax			; cx -= length fitting at end
					; = remaining length to store at start
	xchg cx, ax			; cx = length fitting at end
	call .mov_and_put		; store at end
	xchg cx, ax			; cx = remaining length to store at start

	mov di, word [bp + ?original_dst]
		; literals never go here
		; the following handles source overlapping destination
		;  s.t. source crosses the circular buffer boundary.
	mov ax, si
	add ax, cx			; -> behind source
	jc .part			; CY, must copy part -->
	cmp ax, dx			; below-or-equal end of buffer ?
	jbe .start			; yes -->
.part:
	mov ax, dx
	sub ax, si			; ax = how much source data remains
	sub cx, ax			; cx = how much to copy after
	xchg cx, ax			; cx = part length (source data remains)
	call .mov_and_put		; put from end
	xchg cx, ax			; cx = how much to copy from buffer start
	mov si, word [bp + ?original_dst]	; -> start of buffer
	jmp .mov			; continue with di -> near start of buffer

.start:
	mov di, word [bp + ?original_dst]
.mov:
	call .mov_and_put

.end:
	mov word [bp + ?dst], di
	clc
.ret:
	retn

.ret_CY:
	stc
	retn

.mov_and_put:
	 push ds
	 push si
	push es
	push di
	push cx
	rep movsb
	pop cx
	pop si
	pop ds
	call put_file_data
	 pop si
	 pop ds
	add si, cx
	retn


		; INP:	cx = 0..15
		; OUT:	NC if successful,
		;	 ax = value read
		;	 cx = 0
		;	CY if error
		; CHG:	ds, si
get_bits:
	push bx
	mov bx, word [bp + ?low_bit_index_and_high_current_byte]
	cmp cx, 15
	ja .error
	test cx, cx
	jz .error
	xor ax, ax
.loop:
	test bl, bl
	jnz .havebit
	push ax
	call read_byte
	mov bh, al
	mov bl, 80h
	pop ax
	jc .error
.havebit:
	shl ax, 1
	test bh, bl
	jz @F
	inc ax
@@:
	shr bl, 1
	loop .loop
.end:
	db __TEST_IMM8
.error:
	stc
	mov word [bp + ?low_bit_index_and_high_current_byte], bx
	pop bx
	retn


		; INP:	?src_remaining
		; OUT:	NC if success,
		;	 al = value read
		;	 ?src_remaining decremented
		;	CY if error (source buffer too small),
		;	 ?src_remaining = 0
		; CHG:	ds, si
read_byte:
	sub word [bp + ?src_remaining], 1
	sbb word [bp + ?src_remaining + 2], 0
	jb .empty

	call get_file_byte
	retn

.empty:
	and word [bp + ?src_remaining], 0
	and word [bp + ?src_remaining + 2], 0
	stc
	retn

	lleave ctx

%ifn _STANDALONE
	resetdef COUNTER
	resetdef DEBUG0
%endif
